#include <assert.h>
#include "lib/springspace.h"

//right now only 2 (linear) and 4 (cubic) are supported
#define SPRING_SPACE_DIM	2
#define USE_INNER_CROSS_SPRINGS
#define USE_OUTER_CROSS_SPRINGS

static const int vertexCount = SPRING_SPACE_DIM*SPRING_SPACE_DIM*SPRING_SPACE_DIM;
static const int springCount =
		SPRING_SPACE_DIM*SPRING_SPACE_DIM*(SPRING_SPACE_DIM-1)*3			//box edges
#ifdef USE_INNER_CROSS_SPRINGS
		+(SPRING_SPACE_DIM-1)*(SPRING_SPACE_DIM-1)*(SPRING_SPACE_DIM-1)*4	//inner cross
#endif
#ifdef USE_OUTER_CROSS_SPRINGS
		+(SPRING_SPACE_DIM-1)*(SPRING_SPACE_DIM-1)*SPRING_SPACE_DIM*6		//outer cross
#endif
		;

void SpringSpace::clear() {
	SpringObj::clear();
	model = NULL;
	tvtx.clear();
}

//bool SpringSpace::init(ObjFile *model, ParSystem *parSys, SpringSystem *springSys) {
//	shutdown();	//just in case anything has been loaded
//
//	//notice - when SpringObj::init() first calls shutdown() will it be a call to SpringSpace:: or SpringObj:: ?
//	if (!SpringObj::init(parSys, springSys)) return false;
bool SpringSpace::setModel(ObjFile *model) {
	if (!init(parSys, springSys)) return false;

	this->model = model;
	if (!model) return true;

	const box3f &bbox = model->getBBox();

	int e,i,j,k;

	//vertices
	vertex.resize(vertexCount);
	for (e = k = 0; k < SPRING_SPACE_DIM; k++) {
		float z = (float)k / (float)(SPRING_SPACE_DIM-1) * (bbox.max.z - bbox.min.z) + bbox.min.z;
		for (j = 0; j < SPRING_SPACE_DIM; j++) {
			float y = (float)j / (float)(SPRING_SPACE_DIM-1) * (bbox.max.y - bbox.min.y) + bbox.min.y; 
			for (i = 0; i < SPRING_SPACE_DIM; i++, e++) {
				float x = (float)i / (float)(SPRING_SPACE_DIM-1) * (bbox.max.x - bbox.min.x) + bbox.min.x; 

				vertex[e] = parSys->createParticle();
				if (vertex[e] == NULL) return false;

				Particle *p = vertex[e];
				if (!p) return false;

				p->massInv = 1.f;
				p->x.set(x,y,z);
				p->v.set(0,0,0);
				p->clearForce();
			}
		}
	}

	assert(e == vertex.size());

	//stretched vertices
	tvtx.resize(model->getVertexes().size());

	//springs
	e = 0;
	spring.resize(springCount);
		//edge springs
	for (k = 0; k < SPRING_SPACE_DIM; k++) {
		for (j = 0; j < SPRING_SPACE_DIM; j++) {
			for (i = 0; i < SPRING_SPACE_DIM-1; i++) {
				for (int dim = 0; dim < 3; dim++) {

					int v0, v1;
					switch (dim) {
					case 0:
						v0 = i + j*SPRING_SPACE_DIM + k*SPRING_SPACE_DIM*SPRING_SPACE_DIM;
						v1 = i+1 + j*SPRING_SPACE_DIM + k*SPRING_SPACE_DIM*SPRING_SPACE_DIM;
						break;
					case 1:
						v0 = k + i*SPRING_SPACE_DIM + j*SPRING_SPACE_DIM*SPRING_SPACE_DIM;
						v1 = k + (i+1)*SPRING_SPACE_DIM + j*SPRING_SPACE_DIM*SPRING_SPACE_DIM;
						break;
					case 2:
						v0 = j + k*SPRING_SPACE_DIM + i*SPRING_SPACE_DIM*SPRING_SPACE_DIM;
						v1 = j + k*SPRING_SPACE_DIM + (i+1)*SPRING_SPACE_DIM*SPRING_SPACE_DIM;
						break;
					}

					spring[e] = springSys->createSpring();
					if (spring[e] == -1) return false;
					Spring &s = springSys->spring[spring[e]];
					if (!s.active) return false;
					s.par[0] = vertex[v0];
					s.par[1] = vertex[v1];
					s.restlen = (float)(vertex[v0]->x - vertex[v1]->x).len();
					s.ks = 360;
					s.kd = 3.6f;

					e++;
				}
			}
		}
	}

#ifdef USE_INNER_CROSS_SPRINGS
		//cross springs
	for (k = 0; k < SPRING_SPACE_DIM-1; k++) {
		for (j = 0; j < SPRING_SPACE_DIM-1; j++) {
			for (i = 0; i < SPRING_SPACE_DIM-1; i++) {
				for (int dim = 0; dim < 4; dim++) {

					//use the 'dim' bit vectors to tell which axis to increase along
					//logical-NOT 'dim' to find the opposite corner of the spring
					int v0 = i+(dim&1) + (j+((dim&2)>>1))*SPRING_SPACE_DIM + (k+((dim&4)>>2))*SPRING_SPACE_DIM*SPRING_SPACE_DIM;
					int v1 = i+((~dim)&1) + (j+(((~dim)&2)>>1))*SPRING_SPACE_DIM + (k+(((~dim)&4)>>2))*SPRING_SPACE_DIM*SPRING_SPACE_DIM;

					spring[e] = springSys->createSpring();
					if (spring[e] == -1) return false;
					Spring &s = springSys->spring[spring[e]];
					if (!s.active) return false;
					s.par[0] = vertex[v0];
					s.par[1] = vertex[v1];
					s.restlen = (float)(vertex[v0]->x - vertex[v1]->x).len();
					s.ks = 360;
					s.kd = 3.6f;

					e++;
				}
			}
		}
	}
#endif

#ifdef USE_OUTER_CROSS_SPRINGS
		//outer cross springs
	for (k = 0; k < SPRING_SPACE_DIM-1; k++) {
		for (j = 0; j < SPRING_SPACE_DIM-1; j++) {
			for (i = 0; i < SPRING_SPACE_DIM; i++) {
				for (int dim = 0; dim < 6; dim++) {
					int v0, v1;
					switch (dim) {
					case 0://x axis
						v0 = i + j*SPRING_SPACE_DIM + k*SPRING_SPACE_DIM*SPRING_SPACE_DIM;
						v1 = i + (j+1)*SPRING_SPACE_DIM + (k+1)*SPRING_SPACE_DIM*SPRING_SPACE_DIM;
						break;
					case 1:
						v0 = i + j*SPRING_SPACE_DIM + (k+1)*SPRING_SPACE_DIM*SPRING_SPACE_DIM;
						v1 = i + (j+1)*SPRING_SPACE_DIM + k*SPRING_SPACE_DIM*SPRING_SPACE_DIM;
						break;
					case 2://y axis
						v0 = k + i*SPRING_SPACE_DIM + j*SPRING_SPACE_DIM*SPRING_SPACE_DIM;
						v1 = (k+1) + i*SPRING_SPACE_DIM + (j+1)*SPRING_SPACE_DIM*SPRING_SPACE_DIM;
						break;
					case 3:
						v0 = (k+1) + i*SPRING_SPACE_DIM + j*SPRING_SPACE_DIM*SPRING_SPACE_DIM;
						v1 = k + i*SPRING_SPACE_DIM + (j+1)*SPRING_SPACE_DIM*SPRING_SPACE_DIM;
						break;
					case 4://z axis
						v0 = j + k*SPRING_SPACE_DIM + i*SPRING_SPACE_DIM*SPRING_SPACE_DIM;
						v1 = (j+1) + (k+1)*SPRING_SPACE_DIM + i*SPRING_SPACE_DIM*SPRING_SPACE_DIM;
						break;
					case 5:
						v0 = j + (k+1)*SPRING_SPACE_DIM + i*SPRING_SPACE_DIM*SPRING_SPACE_DIM;
						v1 = (j+1) + k*SPRING_SPACE_DIM + i*SPRING_SPACE_DIM*SPRING_SPACE_DIM;
						break;
					}

					spring[e] = springSys->createSpring();
					if (spring[e] == -1) return false;
					Spring &s = springSys->spring[spring[e]];
					if (!s.active) return false;
					s.par[0] = vertex[v0];
					s.par[1] = vertex[v1];
					s.restlen = (float)(vertex[v0]->x - vertex[v1]->x).len();
					s.ks = 360;
					s.kd = 3.6f;

					e++;
				}
			}
		}
	}
#endif


	assert(e == spring.size());

	return true;
}

#if SPRING_SPACE_DIM != 4 
#if SPRING_SPACE_DIM != 2 
#error all we have implemneted is bernstein coefficients for cubic and linear.  nothing general case yet.
#endif
#endif

#if SPRING_SPACE_DIM == 4
static inline float Bernstein(int i, float t) {
	switch (i) {
	case 0: 
		{
			float s = 1.f - t;
			return s * s * s;
		}
	case 1:
		{
			float s = 1.f - t;
			return 3.f * s * s * t;
		}
	case 2:
		return 3*t*t*(1-t);
	case 3:
		return t*t*t;
	}

	assert(0);
	return 0;
}
#elif SPRING_SPACE_DIM == 2
static inline float Bernstein(int i, float t) {
	switch (i) {
	case 0: return 1.f - t;
	case 1: return t;
	}

	assert(0);
	return 0;
}
#endif

void SpringSpace::render(bool useMtl) {
	//for all vertices
	//calculate their fractional positions in their bbox
	//compute the bezier space position according to the many control points

	const std::vector<vec3f> &mdlVertex = model->getVertexes();
	const box3f &bbox = model->getBBox();
	vec3f invs(
		1.f / (bbox.max.x - bbox.min.x),
		1.f / (bbox.max.y - bbox.min.y),
		1.f / (bbox.max.z - bbox.min.z));

	for (int v = 0; v < (int)mdlVertex.size(); v++) {

		tvtx[v].set(0,0,0);

		//bezier
		int e = 0;
		for (int k = 0; k < SPRING_SPACE_DIM; k++) {
			float bk = Bernstein(k, (mdlVertex[v].z - bbox.min.z) * invs.z);
			for (int j = 0; j < SPRING_SPACE_DIM; j++) {
				float bj = Bernstein(j, (mdlVertex[v].y - bbox.min.y) * invs.y);
				for (int i = 0; i < SPRING_SPACE_DIM; i++) {
					float bi = Bernstein(i, (mdlVertex[v].x - bbox.min.x) * invs.x);
					float b = bi*bj*bk;
					tvtx[v].x += b * vertex[e]->x.x;
					tvtx[v].y += b * vertex[e]->x.y;
					tvtx[v].z += b * vertex[e]->x.z;
					e++;
				}
			}
		}
	}

	glColor3f(1,1,1);
	model->renderImmediate(&tvtx[0], useMtl);
}
